# Copyright (c) 2025 Bytedance Ltd. and/or its affiliates
# SPDX-License-Identifier: MIT

from .template import apply_prompt_template, get_prompt_template

__all__ = [
    "apply_prompt_template",
    "get_prompt_template",
]
